"""MNE visual_92_categories dataset."""

from .visual_92_categories import (data_path, has_visual_92_categories_data,
                                   get_version,
                                   requires_visual_92_categories_data)
